#pragma once
#include "linalg/type.hpp"
#include "pcommon/expected.hpp"

enum class solve_tridiag_error_t {
  UNKNOWN = 0,
  WRONG_L_B = 1,
  WRONG_L_AC = 1,
};

// return a int zero if no error
inline expected<int, solve_tridiag_error_t>
solve_tridiag_thomas(dynmat_t &X, const dynvec_t &a, const dynvec_t &b,
                     const dynvec_t &c) {
  /*
              https://en.wikipedia.org/wiki/Tridiagonal_matrix_algorithm
   solves AX = D, where A is a tridiagonal matrix consisting of vectors a, b, c
               X: matrix initially contains the D, and return X.
   a[] = subdiagonal, indexed from [0, ..., X - 2]
   b[] = main diagonal, indexed from [0, ..., X - 1]
   c[] = superdiagonal, indexed from [0, ..., X - 2]
   */
  dynmat_t S = dynmat_t::Zero(X.rows(), X.cols());
  auto ROWS = X.rows();
  auto ROWS1 = ROWS-1;
  if (b.size() != ROWS) {
    return solve_tridiag_error_t::WRONG_L_B;
  }
	if (a.size() != ROWS1 || c.size() != ROWS1) {
    return solve_tridiag_error_t::WRONG_L_AC;
	}

  for (size_t col = 0; col < S.cols(); ++col) {
    S(0, col) = c[0] / b[0];
    X(0, col) = X(0, col) / b[0];

    /* loop from 1 to X - 1 inclusive */
    for (int ix = 1; ix < ROWS; ix++) {
      S(ix, col) = c[ix] / (b[ix] - a[ix-1] * S(ix - 1, col));
      X(ix, col) = (X(ix, col) - a[ix-1] * X(ix - 1, col)) /
                   (b[ix] - a[ix-1] * S(ix - 1, col));
    }

    /* loop from ROWS - 2 to 0 inclusive */
    for (int ix = ROWS - 2; ix >= 0; ix--)
      X(ix, col) -= S(ix, col) * X(ix + 1, col);
  }
  return 0;
}
