#include <stddef.h>
#include <type_traits>
#include <exception>
#include "pcommon/likely.hpp"
class BadExpectedErrorAccess	 : public std::exception {
public:
  explicit BadExpectedErrorAccess() noexcept = default;
  BadExpectedErrorAccess(BadExpectedErrorAccess const&) {}
  BadExpectedErrorAccess& operator=(BadExpectedErrorAccess const&) { return *this; }
  char const* what() const noexcept override { return "bad expected ErrorAccess"; }
};

class BadExpectedValueAccess	 : public std::exception {
public:
  explicit BadExpectedValueAccess() noexcept = default;
  BadExpectedValueAccess(BadExpectedValueAccess const&) {}
  BadExpectedValueAccess& operator=(BadExpectedValueAccess const&) { return *this; }
  char const* what() const noexcept override { return "bad expected ValueAccess"; }
};

// unexpected is implicitly convertible to expected object in the error state.
template <class Error>
class unexpected final {
  template <class E>
  friend class unexpected;
  template <class V, class E>
  friend class Expected;
 public:
  /**
   * Constructors
   */
  unexpected() = default;
  unexpected(const unexpected&) = default;
  unexpected(unexpected&&) = default;
  unexpected& operator=(const unexpected&) = default;
  unexpected& operator=(unexpected&&) = default;
  explicit unexpected(const Error& err)
      : error_(err) {}
  explicit unexpected(Error&& err)
      : error_(std::move(err)) {}

  template <class Other>
  constexpr /* implicit */ unexpected(unexpected<Other> that)
      : error_(std::move(that.error())) {}

  /**
   * Assignment
   */
  template <class Other>
  unexpected& operator=(unexpected<Other> that) {
    error_ = std::move(that.error());
  }

  /**
   * Observers
   */
  Error& error() & { return error_; }
  const Error& error() const& { return error_; }
  Error&& error() && { return std::move(error_); }

 private:
  Error error_;
};

template <class Error>
inline bool operator==(
    const unexpected<Error>& lhs, const unexpected<Error>& rhs) {
  return lhs.error() == rhs.error();
}

template<typename Value, typename Error>
struct expected {
public:
enum class Which : unsigned char { eEmpty, eValue, eError };
using value_type = Value;
using error_type = Error;
static_assert(
	!std::is_reference<Value>::value,
	"Expected may not be used with reference types");
static_assert(
	!std::is_abstract<Value>::value,
	"Expected may not be used with abstract types");

expected():which_(Which::eEmpty) {}
// explicit expected(const expected& that) = default;
// explicit expected(expected&& that) = default;
/*implicit*/ expected(Value v): value_(v), which_(Which::eValue) {}
// explicit expected(const Value& v): value_(v), which_(Which::eValue) {}
// explicit expected(Value&& v): value_(v), which_(Which::eValue) {}
/*implicit*/ expected(Error v): error_(v), which_(Which::eError) {}
// explicit expected(const Error& v): error_(v), which_(Which::eError) {}
// explicit expected(Error&& v): error_(v), which_(Which::eError) {}
/* implicit */ expected(const unexpected<Error>& err):
error_(err.error()), which_(Which::eError) {}
/* implicit */ expected(unexpected<Error>&& err):
error_(std::move(err.error())), which_(Which::eError) {}

constexpr bool has_value() const noexcept {
	return LIKELY(Which::eValue == this->which_);
}

constexpr bool has_error() const noexcept {
	return UNLIKELY(Which::eError == this->which_);
}

const Value& value() const& {
	requireValue();
	return value_;
}

Value& value() & {
	requireValue();
	return value_;
}

const Error& error() const& {
	requireError();
	return error_;
}

Error& error() & {
	requireError();
	return error_;
}

explicit constexpr operator bool() const noexcept { return has_value(); }

// Return a copy of the value if set, or a given default if not.
template <class U>
Value value_or(U&& dflt) const& {
	if (LIKELY(this->which_ == Which::eValue)) {
		return this->value_;
	}
	return static_cast<U&&>(dflt);
}
private:
void requireValue() const {
	if (UNLIKELY(!has_value())) {
		throw BadExpectedValueAccess();
	}
}

void requireError() const {
	if (UNLIKELY(!has_error())) {
		throw BadExpectedErrorAccess();
	}
}
union {
	Value value_;
	Error error_;
	char ch_;
};
Which which_;
};

template <class Error>
constexpr unexpected<typename std::decay<Error>::type>
make_unexpected(Error&& err) {
  return unexpected<typename std::decay<Error>::type>{
      static_cast<Error&&>(err)};
}

template <class Error, class Value>
constexpr expected<typename std::decay<Value>::type, Error>
make_expected(Value&& val) {
  return expected<typename std::decay<Value>::type, Error>{
      static_cast<Value&&>(val)};
}
