#pragma once
#include <stdexcept>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <execinfo.h>
#include <unistd.h>
#include <limits>

constexpr inline double DNAN = std::numeric_limits<double>::quiet_NaN();
constexpr inline float FNAN = std::numeric_limits<float>::quiet_NaN();

#define _concat(x, y) x##y
#define CONCAT(x, y) _concat(x, y)

#define _loc2(f, l) f ":" #l
#define _loc(f, l) _loc2(f, l)
#define SRC_LOC _loc(__FILE__, __LINE__)


using exit_handler_fn = void (*)(const char*, int, const char*, va_list);

[[noreturn]] static void default_exit_handler(const char* location, int status,
                                              const char* format,
                                              va_list args) {
    ::vfprintf(stderr, format, args);
    va_end(args);
    ::fprintf(stderr, " [@%s]\n", location);
#ifndef _WIN32
    if (status != EXIT_SUCCESS) {
        void* data[512];
        auto size = backtrace(data, sizeof(data) / sizeof(data[0]));
        ::fprintf(stderr, "Backtrace %d:\n", size);
        backtrace_symbols_fd(data, size, STDERR_FILENO);
    }
#endif
    ::exit(status);
}

static auto exit_handler = default_exit_handler;

inline void set_exit_handler(exit_handler_fn handler) {
    exit_handler = handler;
}

[[noreturn, gnu::format(printf, 3, 4)]] inline void __exit_handler(
    const char* location, int status, const char* format, ...) {
    va_list args;
    va_start(args, format);
    exit_handler(location, status, format, args);
    ::exit(status);
}

#ifndef EXIT
#define EXIT(status, fmt, ...)                                     \
    do {                                                           \
        __exit_handler(SRC_LOC, status, "EXIT(" #status "): " fmt, \
                       ##__VA_ARGS__);                             \
    } while (0)
#endif

#ifndef EXIT_OK
#define EXIT_OK(fmt, ...)                                                      \
    do {                                                                       \
        __exit_handler(SRC_LOC, EXIT_SUCCESS, "EXIT_OK: " fmt, ##__VA_ARGS__); \
    } while (0)
#endif

#ifndef EXIT_ERR
#define EXIT_ERR(fmt, ...)                                      \
    do {                                                        \
        __exit_handler(SRC_LOC, EXIT_FAILURE, "EXIT_ERR: " fmt, \
                       ##__VA_ARGS__);                          \
    } while (0)
#endif

#ifndef EXIT_ERRNO
#define EXIT_ERRNO(fn, fmt, ...) \
    EXIT_ERR(fn ": %s -- " fmt, strerror(errno), ##__VA_ARGS__)
#endif

using abort_handler_fn = void (*)(const char*, const char*, va_list);

[[noreturn]] static void default_abort_handler(const char* location,
                                               const char* format,
                                               va_list args) {
    ::fprintf(stderr, "ABORT: ");
    ::vfprintf(stderr, format, args);
    va_end(args);
    ::fprintf(stderr, " [@%s]\n", location);
    ::abort();
}

static auto abort_handler = default_abort_handler;

inline void set_abort_handler(abort_handler_fn handler) {
    abort_handler = handler;
}

[[noreturn, gnu::format(printf, 2, 3)]] inline void __abort_handler(
    const char* location, const char* format, ...) {
    va_list args;
    va_start(args, format);
    abort_handler(location, format, args);
    ::abort();
}

#ifndef ABORT
#define ABORT(fmt, ...)                               \
    do {                                              \
        __abort_handler(SRC_LOC, fmt, ##__VA_ARGS__); \
    } while (0)
#endif

#ifndef ABORT_ERRNO
#define ABORT_ERRNO(fn, fmt, ...) \
    ABORT(fn ": %s -- " fmt, strerror(errno), ##__VA_ARGS__)
#endif

#ifndef EXPECT
#define EXPECT(COND, fmt, ...) \
    if (!(COND)) {\
        EXIT_ERR(#COND " == false: " fmt, ##__VA_ARGS__); \
    }
#endif

namespace common {
class og_exception : public std::runtime_error {
    using std::runtime_error::runtime_error;
};
}

[[noreturn]] static void default_exception(const char* location,
        const char* format, va_list args) {
    char what[4096]{};
    constexpr size_t what_strlen = sizeof(what) - 1;
    size_t offset = 0;
    int ret = ::vsnprintf(what, what_strlen, format, args);
    if (ret < 0) {
        throw ::common::og_exception(what);
    } else {
        offset += ret;
    }
    va_end(args);
    if (offset >= what_strlen) {
        throw ::common::og_exception(what);
    }
    ret = ::snprintf(what+offset, what_strlen - offset, " [@%s]\n", location);
    throw ::common::og_exception(what);
}

[[noreturn, gnu::format(printf, 2, 3)]] inline void __exception_handler(
        const char* location, const char* format, ...) {
    va_list args;
    va_start(args, format);
    default_exception(location, format, args);
}

#ifndef THROW
#define THROW(fmt, ...) \
    do { \
        __exception_handler(SRC_LOC, fmt, ##__VA_ARGS__); \
    } while (0)
#endif
#ifndef THROW_ERRNO
#define THROW_ERRNO(fn, fmt, ...) \
    THROW(fn ": %s -- " fmt, strerror(errno), ##__VA_ARGS__)
#endif

#ifdef USE_OG_EXCEPTION
#undef EXIT_ERR
#define EXIT_ERR THROW

#undef EXIT_ERRNO
#define EXIT_ERRNO THROW_ERRNO
#endif

