#include "linalg/type.hpp"
#include <fit/crs.hpp>
#include <linalg/tridiag.hpp>

using Eigen::ArrayXd;

dynmat_t natural_f(dynvec_t knots) {
  const size_t K = knots.size();
  dynvec_t h = knots.tail(K - 1) - knots.head(K - 1);
  // length = K-2
  dynvec_t diag = ((h.head(K - 2) + h.tail(K - 2))) / 3;
  // length = K-3
  dynvec_t ul_diag = h.segment(1, K - 3) / 6;
  dynmat_t d = dynmat_t::Zero(K - 2, K);
  for (size_t i = 0; i < K - 2; ++i) {
    d(i, i) = 1.0 / h(i);
    d(i, i + 2) = 1.0 / h(i + 1);
    d(i, i + 1) = (-d(i, i)) - d(i, i + 2);
  }
  auto res = solve_tridiag_thomas(d, diag, ul_diag, ul_diag);
  dynmat_t f = dynmat_t::Zero(K, K);
  f.middleRows(1, K - 2) = d;
  return f;
};

std::map<double, int> get_sorted_kidx(const dynvec_t &k) {
  std::map<double, int> sorted_kidx;
  for (size_t i = 0; i < k.size(); ++i) {
    sorted_kidx[k(i)] = i;
  }
  return sorted_kidx;
}

// No assumption on whether knots/x is sorted
Eigen::ArrayXi lower_bound_index(const dynvec_t &x, const dynvec_t &knots) {
  const size_t K = knots.size();
  std::map<double, int> sorted_kidx = get_sorted_kidx(knots);
  Eigen::ArrayXi j = Eigen::ArrayXi::Zero(x.size());
  for (size_t i = 0; i < x.size(); ++i) {
    auto l = sorted_kidx.lower_bound(x(i));
    if (l == sorted_kidx.end()) {
      j(i) = K - 2;
    } else {
      j(i) = std::max(l->second - 1, 0);
    }
  }
  return j;
}

dynmat_t get_free_crs_dmat(dynvec_t x, dynvec_t knots, dynmat_t f,
                           int deriv_order) {
  auto [ajm, ajp, cjm, cjp, j] = compute_base_functions(x, knots, deriv_order);
  Eigen::ArrayXi j1 = j + 1;
  const size_t K = knots.size();
  dynmat_t i = dynmat_t::Identity(K, K);
  using Eigen::all;
  return i(j, all).array().colwise() * ajm + i(j1, all).array() +
         f(j, all).array().colwise() * cjm +
         f(j1, all).array().colwise() * cjp;
}

spline_base_t compute_base_functions(const dynvec_t &x, const dynvec_t &knots,
                                     int deriv_order) {
  // TODO: verify that knots are sorted
  spline_base_t res{};
  res.j = lower_bound_index(x, knots);
  const size_t K = knots.size();
  ArrayXd h = knots.tail(K - 1) - knots.head(K - 1);
  ArrayXd hj = h(res.j);
  ArrayXd xj1_x = knots(res.j + 1) - x;
  ArrayXd x_xj = x - knots(res.j);
  if (deriv_order == 0) {
    res.ajm = xj1_x / hj;
    res.ajp = x_xj / hj;
    ArrayXd cjm_3 = xj1_x.pow(3) / (6. * hj);
    cjm_3 = (x.array() > knots(K - 1)).select(ArrayXd::Zero(x.size()), cjm_3);
    ArrayXd cjm_1 = hj * xj1_x / (6.);
    res.cjm = cjm_3 - cjm_1;
    ArrayXd cjp_3 = x_xj.pow(3) / (6. * hj);
    cjp_3 = (x.array() < knots(0)).select(ArrayXd::Zero(x.size()), cjp_3);
    ArrayXd cjp_1 = hj * xj1_x / (6.);
    res.cjp = cjp_3 - cjp_1;
  } else if (deriv_order == 1) {
    res.ajm = -hj.inverse();
    res.ajp = hj.inverse();
    ArrayXd cjm_3 = -xj1_x.pow(2) / (2. * hj);
    cjm_3 = (x.array() > knots(K - 1)).select(ArrayXd::Zero(x.size()), cjm_3);
    ArrayXd cjm_1 = -hj / (6.);
    res.cjm = cjm_3 - cjm_1;
    ArrayXd cjp_3 = x_xj.pow(2) / (2. * hj);
    cjp_3 = (x.array() < knots(0)).select(ArrayXd::Zero(x.size()), cjp_3);
    ArrayXd cjp_1 = hj / (6.);
    res.cjp = cjp_3 - cjp_1;
  } else if (deriv_order == 2) {
    res.ajm = ArrayXd::Zero(x.size());
    res.ajp = ArrayXd::Zero(x.size());
    ArrayXd cjm_3 = xj1_x / hj;
    cjm_3 = (x.array() > knots(K - 1)).select(ArrayXd::Zero(x.size()), cjm_3);
    ArrayXd cjm_1 = ArrayXd::Zero(x.size());
    res.cjm = cjm_3 - cjm_1;
    ArrayXd cjp_3 = x_xj / hj;
    cjp_3 = (x.array() < knots(0)).select(ArrayXd::Zero(x.size()), cjp_3);
    ArrayXd cjp_1 = ArrayXd::Zero(x.size());
    res.cjp = cjp_3 - cjp_1;
  }
  return res;
}

volfit_iarray_t::volfit_iarray_t(int num_fdata, int num_k):
  num_fitdata(num_fdata),
  num_knots(num_k) {
    fitdata_x = new double[num_fdata]();
    fitdata_values = new double[num_fdata]();
    fitdata_weights = new double[num_fdata]();
    knots = new double[num_knots]();
    x0 = new double[num_knots]();
  }

  volfit_iarray_t::~volfit_iarray_t() {
    if (fitdata_x) {
      delete [] fitdata_x;
    }
    if (fitdata_values) {
      delete [] fitdata_values;
    }
    if (fitdata_weights) {
      delete [] fitdata_weights;
    }
    if (knots) {
      delete [] knots;
    }
    if (x0) {
      delete [] x0;
    }
  }
