#include <math.h>
#include <float.h>
#include <vector>
#include <time.h>
#include <stdio.h>

#include "lets_be_rational.h"

#if defined(_MSC_VER)
# pragma warning (disable : 26451)
#endif

#include <cstring> // for strrchr()

namespace { const char* skip_path(const char* file_name) { return (const char*)std::max((uintptr_t)file_name, std::max((uintptr_t)strrchr(file_name, '/'), (uintptr_t)strrchr(file_name, '\\')) + 1); } }

// Compile with:
//                f=lets_be_rational_timing_test; g++ -DNDEBUG -Ofast $f.cpp -o Linux/$f -s -Wl,-rpath=. -L./Linux -l:LetsBeRational.so
// or
//                f=lets_be_rational_timing_test; g++ -DNDEBUG -Ofast $f.cpp -o x64/$f   -s -Wl,-rpath=. -L./Linux -l:LetsBeRational.xll

//
//  x = ln(F/K)
//
//  β = price / √(F·K)
//
//  Thus, when F = 1, we have K = exp(-x) and
//
//     β = price · exp(x/2)
//     price = β · exp(-x/2)
//     price = β · √K
//
// Invocation example:
//
//   ./lets_be_rational_timing_test  1048576  0 0.00000001 0.0001 0.001 0.01 0.1 0.25 0.5 1 2 4 8 16 32 64 128 256 512
//

#define NO_PREMAPPING_OF_VOLATILITIES

int main(int argc, char** argv) {
  if (argc < 3) {
    fprintf(stderr, "Usage: %s <number_of_points> <x> [ <x2> <x3> ...]\n", skip_path(argv[0]));
    fprintf(stderr, "\nInvocation example:\n\n   %s   1048576  0 0.00000001 0.0001 0.001 0.01 0.1 0.25 0.5 1 2 4 8 16 32 64 128 256 512\n\n", skip_path(argv[0]));
    fflush(stderr);
    return -1;
  }
  const int m = std::max((int)strtod(argv[1], 0), 4), m1 = m / 2;
  std::vector<double> x_values(argc - 2), s(m), price(m), sigma_input(m), sigma_output(m);
  for (int i_arg = 2; i_arg < argc; ++i_arg) x_values[(size_t)i_arg - 2] = strtod(argv[i_arg], 0);
  printf("\n%s\t%d\t%g", skip_path(argv[0]), m, x_values[0]);
  for (int i_arg = 3; i_arg < argc; ++i_arg) printf(" %g", x_values[(size_t)i_arg - 2]);
  printf("\n");
  printf("\nLoaded %s revision %d [%d bit %s running on %s] compiled with %s on %s from %s .\n\n", DLLName(), Revision(), Bitness(), BuildConfiguration(), CPUName(), CompilerVersion(), BuildDate(), DLLDirectory());
  fflush(stdout);
  double total_elapsed_time = 0, total_rel_sum = 0;
  size_t total_evaluation_count = 0, total_success_count = 0;
  for (const auto x_val : x_values) {
    const double x = -fabs(x_val), F = 1, sqrt_F = sqrt(F), K = F * exp(-x), sqrt_K = sqrt(K), T = 1, sqrt_T = sqrt(T), call = 1;
    const double beta_min = DBL_MIN, s_mid = sqrt(std::max(2 * fabs(x), DBL_EPSILON)), beta_mid = NormalisedBlack(x, s_mid, call), beta_max = exp(x / 2) * (1 - DBL_EPSILON), max_price = beta_max * (sqrt_F * sqrt_K);
    const double s_max = ImpliedBlackVolatility(max_price, F, K, T, call) * sqrt_T, log_beta_min = log(beta_min), d_logb = (log(beta_mid) - log_beta_min) / (m1 - 1), ds = (s_max - s_mid) / (m - m1);
    for (int i = 0; i < m1; ++i)
      s[i] = ImpliedBlackVolatility(exp(log_beta_min + i * d_logb) * (sqrt_F * sqrt_K), F, K, T, call) * sqrt_T;
    for (int i = m1; i < m; ++i)
      s[i] = s_mid + (i - m1 + 1) * ds;
    for (int i = 0; i < m; ++i) {
      price[i] = Black(F, K, sigma_input[i] = s[i] / sqrt_T, T, call);
#if !defined( NO_PREMAPPING_OF_VOLATILITIES )
      sigma_input[i] = ImpliedBlackVolatility(price[i], F, K, T, call);
#endif
    }
    printf("#\n# |x|=%.16g\tnumber_of_points=%d\tbeta_min=%g\tbeta_mid=%g\tbeta_max=%g\ts_min=%g\ts_mid=%g\ts_max=%g\n", fabs(x), m, beta_min, beta_mid, beta_max, s[0], s[(size_t)m1 - 1], s[(size_t)m - 1]);
    fflush(stdout);
    int ibegin, iend;
    for (ibegin = 0; ibegin < m && (price[ibegin] <= 0 || sigma_input[ibegin] <= 0); ++ibegin);
    for (iend = m - 1; iend >= 0 && (price[iend] >= F || sigma_input[iend] >= DBL_MAX); --iend);

    const clock_t start_clock = clock();
    for (int i = ibegin; i <= iend; ++i)
      sigma_output[i] = ImpliedBlackVolatility(price[i], F, K, T, call);
    const clock_t end_clock = clock();

    double sum = 0, rel_sum = 0;
    int evaluation_count = 0, success_count = 0;
    for (int i = ibegin; i <= iend; ++i) {
      ++evaluation_count;
      if (sigma_output[i] > -DBL_MAX && sigma_output[i] < DBL_MAX) {
        ++success_count;
        sum += sigma_output[i];
        double rel_diff = fabs(sigma_output[i] / sigma_input[i] - 1);
#if defined( NO_PREMAPPING_OF_VOLATILITIES ) // Express relative (residual) difference as a proportion of what is theoretically attainable. Numbers lower than 1 suggest that the numerics is, net, more accurate than DBL_EPSILON, which is the nominal granularity.
        rel_diff /= (DBL_EPSILON * (1 + fabs(1 / BlackAccuracyFactor(x, sigma_input[i] * sqrt_T, 1))));
#endif
        rel_sum += rel_diff;
      } else {
        printf("ImpliedBlackVolatility(%g,%g,%g,%g,%g) returned %g.\n", price[i], F, K, T, call, sigma_output[i]);
        fflush(stdout);
      }
    }
    double elapsed_time = (end_clock - start_clock) / ((double)CLOCKS_PER_SEC) * 1000000;
#if defined( NO_PREMAPPING_OF_VOLATILITIES )
    //
    // NOTE: when this number is below 1, the algorithm does better than what is theoretically attainable according to error propagation analysis (1st order).
    // See lets_be_rational.cpp for a derivation.
    //
#define RELATIVE_ACCURACY_OUTPUT "|(attained accuracy)/(theoretically attainable accuracy)|"
#else
#define RELATIVE_ACCURACY_OUTPUT "|(relative volatility difference)|"
#endif
    printf("#\n# Total implied volatility evaluations: %d\tAverage time: %g microseconds\tAverage volatility [%d]: %g\tAverage " RELATIVE_ACCURACY_OUTPUT " : %g\n", evaluation_count, elapsed_time / evaluation_count, success_count, sum / success_count, rel_sum / success_count);
    fflush(stdout);
    total_elapsed_time += elapsed_time;
    total_evaluation_count += evaluation_count;
    total_rel_sum += rel_sum;
    total_success_count += success_count;
  }
  printf("\nAverage over %zu different values for x:  %g  microseconds per evaluation (%zu successful evaluations in total).\n\nTotal average " RELATIVE_ACCURACY_OUTPUT " : %g.\n", x_values.size(), total_elapsed_time / total_evaluation_count, total_evaluation_count, total_rel_sum / total_success_count);
#if defined( NO_PREMAPPING_OF_VOLATILITIES )
  printf("This means the method is on average %g times more accurate than what one could\nhope for on the basis of first order error propagation analysis.\n", floor(total_success_count / total_rel_sum * 10 + 0.5) / 10);
#endif
  printf("\n");
  fflush(stdout);
  return 0;
}

/*

Output on 12th Gen Intel(R) Core(TM) i5-12500H under Windows 11 with Visual-Studio-2022-generated 64 bit LetsBeRational.xll:

    letsberational_timing.exe       1048576 0 1e-08 0.0001 0.001 0.01 0.1 0.25 0.5 1 2 4 8 16 32 64 128 256 512

    Loaded LetsBeRational.xll revision 1515 [64 bit RELEASE running on 12th Gen Intel(R) Core(TM) i5-12500H] compiled with MSVC 1936 on Tue Dec 19 14:43:21 2023 from .\build\x64\Release .

    #
    # |x|=0 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=5.9447e-09     beta_max=1      s_min=5.57743e-308      s_mid=1.49012e-08       s_max=16.4191
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.161171 microseconds     Average volatility [1048576]: 4.10487   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.126921
    #
    # |x|=1e-08     number_of_points=1048576        beta_min=2.22507e-308   beta_mid=5.6414e-05     beta_max=1      s_min=2.71516e-10       s_mid=0.000141421       s_max=16.4191
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.319481 microseconds     Average volatility [1048576]: 4.10491   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.19772
    #
    # |x|=0.0001    number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.00559199     beta_max=0.99995        s_min=2.69694e-06       s_mid=0.0141421 s_max=16.4191
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.338554 microseconds     Average volatility [1048576]: 4.10842   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.204324
    #
    # |x|=0.001     number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.0173442      beta_max=0.9995 s_min=2.69245e-05       s_mid=0.0447214 s_max=16.5847
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.342369 microseconds     Average volatility [1048576]: 4.15711   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.20934
    #
    # |x|=0.01      number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.0515133      beta_max=0.995012       s_min=0.000268798       s_mid=0.141421  s_max=16.4191
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.338554 microseconds     Average volatility [1048576]: 4.1406    Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.194162
    #
    # |x|=0.1       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.13147        beta_max=0.951229       s_min=0.00268352        s_mid=0.447214  s_max=16.4191
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.332832 microseconds     Average volatility [1048576]: 4.21972   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.211347
    #
    # |x|=0.25      number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.169576       beta_max=0.882497       s_min=0.0067044 s_mid=0.707107  s_max=16.419
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.333786 microseconds     Average volatility [1048576]: 4.28894   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.218821
    #
    # |x|=0.5       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.185683       beta_max=0.778801       s_min=0.0134021 s_mid=1 s_max=16.419
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.337601 microseconds     Average volatility [1048576]: 4.3691    Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.225452
    #
    # |x|=1 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.173594       beta_max=0.606531       s_min=0.026791  s_mid=1.41421   s_max=16.5843
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.337601 microseconds     Average volatility [1048576]: 4.52733   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.246053
    #
    # |x|=2 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.122098       beta_max=0.367879       s_min=0.0535553 s_mid=2 s_max=16.7471
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.333786 microseconds     Average volatility [1048576]: 4.74106   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.223734
    #
    # |x|=4 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.0503856      beta_max=0.135335       s_min=0.107058  s_mid=2.82843   s_max=16.9049
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.330925 microseconds     Average volatility [1048576]: 5.03893   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.240832
    #
    # |x|=8 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.00742863     beta_max=0.0183156      s_min=0.21401   s_mid=4 s_max=17.3664
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.33474 microseconds      Average volatility [1048576]: 5.54699   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.223857
    #
    # |x|=16        number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.000144752    beta_max=0.000335463    s_min=0.427814  s_mid=5.65685   s_max=18.2291
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.335693 microseconds     Average volatility [1048576]: 6.36869   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.218832
    #
    # |x|=32        number_of_points=1048576        beta_min=2.22507e-308   beta_mid=5.07396e-08    beta_max=1.12535e-07    s_min=0.855246  s_mid=8 s_max=19.6294
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.33474 microseconds      Average volatility [1048576]: 7.66902   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.205486
    #
    # |x|=64        number_of_points=1048576        beta_min=2.22507e-308   beta_mid=5.88893e-15    beta_max=1.26642e-14    s_min=1.70998   s_mid=11.3137   s_max=22.1223
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.336647 microseconds     Average volatility [1048576]: 9.80433   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.23532
    #
    # |x|=128       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=7.62071e-29    beta_max=1.60381e-28    s_min=3.42094   s_mid=16        s_max=26.097
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.336647 microseconds     Average volatility [1048576]: 13.2324   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.24974
    #
    # |x|=256       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=1.24084e-56    beta_max=2.57221e-56    s_min=6.8601    s_mid=22.6274   s_max=32.2536
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.344276 microseconds     Average volatility [1048576]: 18.7259   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.296774
    #
    # |x|=512       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=3.22573e-112   beta_max=6.61626e-112   s_min=13.8959   s_mid=32        s_max=41.1406
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.344276 microseconds     Average volatility [1048576]: 27.4546   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.316692

    Average over 18 different values for x:  0.326316  microseconds per evaluation (18874368 successful evaluations in total).

    Total average |(attained accuracy)/(theoretically attainable accuracy)| : 0.224745.
    This means the method is on average 4.4 times more accurate than what one could
    hope for on the basis of first order error propagation analysis.

Same output on the same 12th Gen Intel(R) Core(TM) i5-12500H under Windows-Subsystem-for-Linux ('WSL') with g++-generated LetsBeRational.so:

    letsberational_timing   1048576 0 1e-08 0.0001 0.001 0.01 0.1 0.25 0.5 1 2 4 8 16 32 64 128 256 512

    Loaded LetsBeRational.so revision 1515 [64 bit RELEASE running on 12th Gen Intel(R) Core(TM) i5-12500H] compiled with GCC 11.4.0 on Dec 19 2023 14:46:16 from ./Linux/. .

    #
    # |x|=0 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=5.9447e-09     beta_max=1      s_min=5.57743e-308      s_mid=1.49012e-08       s_max=16.4191
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.136059 microseconds     Average volatility [1048576]: 4.10487   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.130939
    #
    # |x|=1e-08     number_of_points=1048576        beta_min=2.22507e-308   beta_mid=5.6414e-05     beta_max=1      s_min=2.71516e-10       s_mid=0.000141421       s_max=16.4191
    #
    # Total implied volatility evaluations: 1048575 Average time: 0.26586 microseconds      Average volatility [1048575]: 4.10491   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.187812
    #
    # |x|=0.0001    number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.00559199     beta_max=0.99995        s_min=2.69694e-06       s_mid=0.0141421 s_max=16.4191
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.27592 microseconds      Average volatility [1048576]: 4.10842   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.192557
    #
    # |x|=0.001     number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.0173442      beta_max=0.9995 s_min=2.69245e-05       s_mid=0.0447214 s_max=16.5847
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.276002 microseconds     Average volatility [1048576]: 4.15712   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.197443
    #
    # |x|=0.01      number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.0515133      beta_max=0.995012       s_min=0.000268798       s_mid=0.141421  s_max=16.4191
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.27792 microseconds      Average volatility [1048576]: 4.14062   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.177351
    #
    # |x|=0.1       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.13147        beta_max=0.951229       s_min=0.00268352        s_mid=0.447214  s_max=16.4191
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.266069 microseconds     Average volatility [1048576]: 4.21981   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.193026
    #
    # |x|=0.25      number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.169576       beta_max=0.882497       s_min=0.0067044 s_mid=0.707107  s_max=16.419
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.270172 microseconds     Average volatility [1048576]: 4.28906   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.201246
    #
    # |x|=0.5       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.185683       beta_max=0.778801       s_min=0.0134021 s_mid=1 s_max=16.419
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.273043 microseconds     Average volatility [1048576]: 4.36918   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.204426
    #
    # |x|=1 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.173594       beta_max=0.606531       s_min=0.026791  s_mid=1.41421   s_max=16.5843
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.268358 microseconds     Average volatility [1048576]: 4.52797   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.217799
    #
    # |x|=2 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.122098       beta_max=0.367879       s_min=0.0535553 s_mid=2 s_max=16.7471
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.270665 microseconds     Average volatility [1048576]: 4.74134   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.200757
    #
    # |x|=4 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.0503856      beta_max=0.135335       s_min=0.107058  s_mid=2.82843   s_max=16.9049
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.269524 microseconds     Average volatility [1048576]: 5.03964   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.209741
    #
    # |x|=8 number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.00742863     beta_max=0.0183156      s_min=0.21401   s_mid=4 s_max=17.3664
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.271397 microseconds     Average volatility [1048576]: 5.54767   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.195603
    #
    # |x|=16        number_of_points=1048576        beta_min=2.22507e-308   beta_mid=0.000144752    beta_max=0.000335463    s_min=0.427814  s_mid=5.65685   s_max=18.2291
    #
    # Total implied volatility evaluations: 1048575 Average time: 0.281742 microseconds     Average volatility [1048575]: 6.36913   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.191726
    #
    # |x|=32        number_of_points=1048576        beta_min=2.22507e-308   beta_mid=5.07396e-08    beta_max=1.12535e-07    s_min=0.855246  s_mid=8 s_max=19.6294
    #
    # Total implied volatility evaluations: 1048575 Average time: 0.276502 microseconds     Average volatility [1048575]: 7.66911   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.187616
    #
    # |x|=64        number_of_points=1048576        beta_min=2.22507e-308   beta_mid=5.88893e-15    beta_max=1.26642e-14    s_min=1.70998   s_mid=11.3137   s_max=22.1223
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.274366 microseconds     Average volatility [1048576]: 9.80438   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.205784
    #
    # |x|=128       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=7.62071e-29    beta_max=1.60381e-28    s_min=3.42094   s_mid=16        s_max=26.097
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.276413 microseconds     Average volatility [1048576]: 13.2325   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.218257
    #
    # |x|=256       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=1.24084e-56    beta_max=2.57221e-56    s_min=6.8601    s_mid=22.6274   s_max=32.2536
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.287099 microseconds     Average volatility [1048576]: 18.7257   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.252023
    #
    # |x|=512       number_of_points=1048576        beta_min=2.22507e-308   beta_mid=3.22573e-112   beta_max=6.61626e-112   s_min=13.8959   s_mid=32        s_max=41.1406
    #
    # Total implied volatility evaluations: 1048576 Average time: 0.270843 microseconds     Average volatility [1048576]: 27.4546   Average |(attained accuracy)/(theoretically attainable accuracy)| : 0.285038

    Average over 18 different values for x:  0.265997  microseconds per evaluation (18874365 successful evaluations in total).

    Total average |(attained accuracy)/(theoretically attainable accuracy)| : 0.20273.
    This means the method is on average 4.9 times more accurate than what one could
    hope for on the basis of first order error propagation analysis.

*/
